#ifdef __cplusplus
extern "C" {
#endif

#include <string.h>
#include <errno.h>
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "br_cmd.h"

#ifdef __cplusplus
}
#endif

#define DEV_BLKSIZE 5    /*  How many devices in a command allocated at a time */
#define ISDIMCMD(cmd) ((cmd == DIM) || (cmd == BRIGHT))


typedef struct {
    int numdevs;
    int allocateddevs;
    int *devs;
} br_device_list;


/*
 *
 * BottleRocket.xs -- Perl module to use BottleRocket commands
 *  (c) 1999 Tymm Twillman.
 *  Not for use in life-saving devices.  This stuff won't get
 *  you dates.  Trust me.
 *
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

int br_getcmd(char *arg)
{
/*
 * Convert a BottleRocket command to the appropriate token
 */

    if (!strcasecmp(arg, "ON"))
        return ON;

    if (!strcasecmp(arg, "OFF"))
        return OFF;

    if (!strcasecmp(arg, "DIM"))
        return DIM;

    if (!strcasecmp(arg, "BRIGHT"))
        return BRIGHT;

    if (!strcasecmp(arg, "ALL_ON"))
        return ALL_ON;

    if (!strcasecmp(arg, "ALL_OFF"))
        return ALL_OFF;

    if (!strcasecmp(arg, "LAMPS_ON"))
        return ALL_LAMPS_ON;

    if (!strcasecmp(arg, "LAMPS_OFF"))
        return ALL_LAMPS_OFF;

    return -1;
}

void my_error_handler(char *where, char *problem)
{
    croak("where->%s  problem->%s  error->%s", where, problem, strerror(errno));
}

int br_strtohc(char *hcptr, char **endptr)
{
    char *my_endptr = hcptr;
    char c;


    while (isspace(*my_endptr))
        my_endptr++;

    if (!*my_endptr) {
        *endptr = hcptr;
        return 0;
    }

    c = toupper(*my_endptr) - 'A';

    if ((c < 0) || (c > 15)) {
        *endptr = hcptr;
        return 0;
    }

    *endptr = ++my_endptr;

    return c;
}

br_device_list *br_new_device_list()
{
    br_device_list *devs;

    devs = malloc(sizeof(br_device_list));

    if (devs == NULL) {
        brerror("br_new_device_list", "malloc");
        return NULL;
    }

#ifdef MEM_DEBUG
    printf("br_new_device_list: Malloced %d bytes at %lx\n",
      sizeof(br_device_list), (unsigned long)devs);
#endif

    devs->allocateddevs = 0;
    devs->numdevs = 0;
    devs->devs = NULL;

    return devs;
}

int br_free_device_list(br_device_list *devs)
{
    if ((devs != NULL) && (devs->devs != NULL)) {

#ifdef MEM_DEBUG
        printf("br_free_device_list: Freeing memory at %lx\n",
          (unsigned long)devs->devs);
        printf("br_free_device_list: Freeing memory at %lx\n",
          (unsigned long)devs);
#endif

        free(devs->devs);
        free(devs);
    }

    return 0;
}

int br_strtodl(char *dlptr, br_device_list *devs, char **endptr)
{
    int dev;
    char *my_endptr = NULL;


    if (devs == NULL) {
        errno = EINVAL;
        brerror("br_strtodl", "NULL device list");
        return -1;
    }

    /* Get rid of any residue */

    if (devs->devs)
        free(devs->devs);

    devs->devs = NULL;
    devs->allocateddevs = 0;
    devs->numdevs = 0;

    do {
        dev = strtol(dlptr, &my_endptr, 0);

        if ((dev > 16) || (dev < 1))
        {
            errno = EINVAL;
            brerror("br_strtodl", "Bad device number");
            return -1;
        }

        if ((*my_endptr != '\0') && (*my_endptr != ',')) {
            *endptr = my_endptr;
            return 0;
        }

        if (br_add_device(devs, dev - 1) < 0)
            return -1;

        dlptr = my_endptr;
    } while (*dlptr++);

    *endptr = --dlptr;

    return 0;
}

int br_add_device(br_device_list *devs, int dev)
{
    if (devs == NULL) {
        errno = EINVAL;
        brerror("br_add_device", "NULL device list");
        return -1;
    }

    if (devs->numdevs >= devs->allocateddevs) {

#ifdef MEM_DEBUG
     printf("br_add_device: Reallocing memory from %lx...\n",
      (unsigned long)devs->devs);
#endif

        devs->devs = realloc(devs->devs, (devs->allocateddevs + DEV_BLKSIZE) * sizeof(int));

        if (devs->devs == NULL) {
            brerror("br_add_device", "realloc");
            return -1;
        }

#ifdef MEM_DEBUG
    printf("br_add_device: Realloced %d bytes at %lx\n",
      (devs->numdevs + DEV_BLKSIZE) * sizeof(int), (unsigned long)devs->devs);
#endif

        devs->allocateddevs += DEV_BLKSIZE;

    }

    devs->devs[devs->numdevs] = dev;

    devs->numdevs++;

    return 0;
}

int br_get_next_device(br_device_list *devs, int *index)
{
    if ((devs == NULL) || (index == NULL))
        return -1;

    if (*index >= devs->numdevs)
        return -1;

    return devs->devs[(*index)++];
}

int br_get_num_devices(br_device_list *devs)
{
    if (devs == NULL)
        return 0;

    return devs->numdevs;
}

MODULE = BottleRocket        PACKAGE = BottleRocket        

void
x10_cmd(FileHandle, Unit, Command, ...)
    FILE *FileHandle
    char *Unit
    char *Command

    PREINIT:
    va_list ap;
    register int i;
    char unit_out;
    int cmd_out;
    int fd;
    char *end;
    br_device_list *devs;
    int housecode;
    int dev;
    int dimlevel;
    int idx = 0;

    CODE:

    cmd_out = br_getcmd(Command);
        
    if (cmd_out < 0) {
        croak("Invalid command");
        return;
    }

    if (ISDIMCMD(cmd_out)) {
            dimlevel = SvIV(ST(3));
    }

    br_error_handler = my_error_handler;
    
    if ((FileHandle == NULL) || ((fd = fileno(FileHandle)) < 0)) {
        croak("Invalid filehandle");
        return;
    }
            
    housecode = br_strtohc(Unit, &end);
    
    if (housecode < 0) {
        croak("Invalid device list");
        return;
    }

    devs = br_new_device_list();
        
    if (*end) {
        br_strtodl(end, devs, &end);
        if (*end != '\0') {
            croak("Invalid device list");
            return;
        }
    } else if ((cmd_out == ON) || (cmd_out == OFF)) {
        croak("No devices specified for command that needs them");
        return;
    }

    if (br_get_num_devices(devs)) {
        while ((dev = br_get_next_device(devs, &idx)) >= 0) {
            if (ISDIMCMD(cmd_out)) {
                x10_br_out(fd, dev, ON);
                for (i = 0; i < dimlevel; i++)
                    x10_br_out(fd, 0, cmd_out);
            } else {
                x10_br_out(fd, dev, cmd_out);
            }
        }
    } else {
        x10_br_out(fd, 0, cmd_out);
    }
    
    br_free_device_list(devs);
        
    return;
